import pytest
import uuid

from pulp_ansible.tests.functional.constants import ANSIBLE_GALAXY_URL
from pulp_ansible.tests.functional.utils import gen_ansible_remote

from pulpcore.client.pulp_ansible import ApiException, AsyncOperationResponse


# TODO Find a way to make the factories work with try_action
def gen_repo():
    return {"name": str(uuid.uuid4())}


def gen_distribution():
    return {"name": str(uuid.uuid4()), "base_path": str(uuid.uuid4())}


@pytest.fixture()
def gen_users(gen_user):
    """Returns a user generator function for the tests."""

    def _gen_users(role_names=list()):
        if isinstance(role_names, str):
            role_names = [role_names]
        creator_roles = [f"ansible.{role}_creator" for role in role_names]
        viewer_roles = [f"ansible.{role}_viewer" for role in role_names]
        user_creator = gen_user(model_roles=creator_roles)
        user_reader = gen_user(model_roles=viewer_roles)
        user_helpless = gen_user()
        return user_creator, user_reader, user_helpless

    return _gen_users


@pytest.fixture()
def try_action(monitor_task):
    def _try_action(user, client, action, outcome, *args, **kwargs):
        action_api = getattr(client, f"{action}_with_http_info")
        try:
            with user:
                response, status, _ = action_api(*args, **kwargs, _return_http_data_only=False)
            if isinstance(response, AsyncOperationResponse):
                response = monitor_task(response.task)
        except ApiException as e:
            assert e.status == outcome, f"{e}"
        else:
            assert status == outcome, f"User performed {action} when they shouldn't been able to"
            return response

    return _try_action


def test_ansible_repository_rbac(ansible_bindings, gen_users, try_action):
    user_creator, user_reader, user_helpless = gen_users("ansiblerepository")

    # List testing
    try_action(user_creator, ansible_bindings.RepositoriesAnsibleApi, "list", 200)
    try_action(user_reader, ansible_bindings.RepositoriesAnsibleApi, "list", 200)
    try_action(user_helpless, ansible_bindings.RepositoriesAnsibleApi, "list", 200)

    # Create testing
    repo = try_action(
        user_creator, ansible_bindings.RepositoriesAnsibleApi, "create", 201, gen_repo()
    )
    try_action(user_reader, ansible_bindings.RepositoriesAnsibleApi, "create", 403, gen_repo())
    try_action(user_helpless, ansible_bindings.RepositoriesAnsibleApi, "create", 403, gen_repo())

    # View testing
    try_action(user_creator, ansible_bindings.RepositoriesAnsibleApi, "read", 200, repo.pulp_href)
    try_action(user_reader, ansible_bindings.RepositoriesAnsibleApi, "read", 200, repo.pulp_href)
    try_action(user_helpless, ansible_bindings.RepositoriesAnsibleApi, "read", 404, repo.pulp_href)

    # Update testing
    update_args = [repo.pulp_href, gen_repo()]
    try_action(user_creator, ansible_bindings.RepositoriesAnsibleApi, "update", 202, *update_args)
    try_action(user_reader, ansible_bindings.RepositoriesAnsibleApi, "update", 403, *update_args)
    try_action(user_helpless, ansible_bindings.RepositoriesAnsibleApi, "update", 404, *update_args)

    # Partial update testing
    partial_update_args = [repo.pulp_href, {"name": str(uuid.uuid4())}]
    try_action(
        user_creator,
        ansible_bindings.RepositoriesAnsibleApi,
        "partial_update",
        202,
        *partial_update_args,
    )
    try_action(
        user_reader,
        ansible_bindings.RepositoriesAnsibleApi,
        "partial_update",
        403,
        *partial_update_args,
    )
    try_action(
        user_helpless,
        ansible_bindings.RepositoriesAnsibleApi,
        "partial_update",
        404,
        *partial_update_args,
    )

    # Delete testing
    try_action(user_reader, ansible_bindings.RepositoriesAnsibleApi, "delete", 403, repo.pulp_href)
    try_action(
        user_helpless, ansible_bindings.RepositoriesAnsibleApi, "delete", 404, repo.pulp_href
    )
    try_action(user_creator, ansible_bindings.RepositoriesAnsibleApi, "delete", 202, repo.pulp_href)


def test_ansible_repository_version_repair(
    ansible_bindings,
    ansible_repo_factory,
    gen_users,
    try_action,
):
    """Test the repository version repair action"""
    user_creator, user_reader, user_helpless = gen_users("ansiblerepository")

    with user_creator:
        repo = ansible_repo_factory()
        ver_href = repo.latest_version_href
    body = {"verify_checksums": True}
    try_action(
        user_creator, ansible_bindings.RepositoriesAnsibleVersionsApi, "repair", 202, ver_href, body
    )
    try_action(
        user_reader, ansible_bindings.RepositoriesAnsibleVersionsApi, "repair", 403, ver_href, body
    )
    try_action(
        user_helpless,
        ansible_bindings.RepositoriesAnsibleVersionsApi,
        "repair",
        403,
        ver_href,
        body,
    )


def test_repository_apis(
    ansible_bindings,
    ansible_collection_remote_factory,
    ansible_repo_factory,
    gen_users,
    try_action,
):
    """Test repository specific actions, modify, sync and rebuild_metadata."""
    user_creator, user_reader, user_helpless = gen_users(["ansiblerepository", "collectionremote"])

    # Sync tests
    with user_creator:
        user_creator_remote = ansible_collection_remote_factory(
            url="https://galaxy.ansible.com",
            requirements_file="collections:\n  - name: community.docker\n    version: 3.0.0",
        )
        repo = ansible_repo_factory()

    body = {"mirror": False, "optimize": True, "remote": user_creator_remote.pulp_href}
    try_action(
        user_reader, ansible_bindings.RepositoriesAnsibleApi, "sync", 403, repo.pulp_href, body
    )
    try_action(
        user_creator, ansible_bindings.RepositoriesAnsibleApi, "sync", 202, repo.pulp_href, body
    )
    try_action(
        user_helpless, ansible_bindings.RepositoriesAnsibleApi, "sync", 404, repo.pulp_href, body
    )

    # Modify tests
    try_action(
        user_reader, ansible_bindings.RepositoriesAnsibleApi, "modify", 403, repo.pulp_href, {}
    )
    try_action(
        user_creator, ansible_bindings.RepositoriesAnsibleApi, "modify", 202, repo.pulp_href, {}
    )
    try_action(
        user_helpless, ansible_bindings.RepositoriesAnsibleApi, "modify", 404, repo.pulp_href, {}
    )

    # Rebuild metadata tests
    rebuild_metadata_args = [
        repo.pulp_href,
        {"namespace": "community", "name": "docker", "version": "3.0.0"},
    ]
    try_action(
        user_reader,
        ansible_bindings.RepositoriesAnsibleApi,
        "rebuild_metadata",
        403,
        *rebuild_metadata_args,
    )
    try_action(
        user_creator,
        ansible_bindings.RepositoriesAnsibleApi,
        "rebuild_metadata",
        202,
        *rebuild_metadata_args,
    )
    try_action(
        user_helpless,
        ansible_bindings.RepositoriesAnsibleApi,
        "rebuild_metadata",
        404,
        *rebuild_metadata_args,
    )


def test_repository_role_management(ansible_bindings, gen_users, ansible_repo_factory, try_action):
    """Check repository role management apis."""
    user_creator, user_reader, user_helpless = gen_users("ansiblerepository")

    with user_creator:
        repo = ansible_repo_factory()
        href = repo.pulp_href

    # Permission check testing
    aperm_response = try_action(
        user_reader, ansible_bindings.RepositoriesAnsibleApi, "my_permissions", 200, href
    )
    assert aperm_response.permissions == []
    bperm_response = try_action(
        user_creator, ansible_bindings.RepositoriesAnsibleApi, "my_permissions", 200, href
    )
    assert len(bperm_response.permissions) > 0
    try_action(user_helpless, ansible_bindings.RepositoriesAnsibleApi, "my_permissions", 404, href)

    # Add "creator" role testing
    nested_role = {"users": [user_helpless.username], "role": "ansible.ansiblerepository_viewer"}
    try_action(
        user_reader,
        ansible_bindings.RepositoriesAnsibleApi,
        "add_role",
        403,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_helpless,
        ansible_bindings.RepositoriesAnsibleApi,
        "add_role",
        404,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_creator,
        ansible_bindings.RepositoriesAnsibleApi,
        "add_role",
        201,
        href,
        nested_role=nested_role,
    )

    # Permission check testing again
    cperm_response = try_action(
        user_helpless, ansible_bindings.RepositoriesAnsibleApi, "my_permissions", 200, href
    )
    assert len(cperm_response.permissions) == 1

    # Remove "viewer" role testing
    try_action(
        user_reader,
        ansible_bindings.RepositoriesAnsibleApi,
        "remove_role",
        403,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_helpless,
        ansible_bindings.RepositoriesAnsibleApi,
        "remove_role",
        403,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_creator,
        ansible_bindings.RepositoriesAnsibleApi,
        "remove_role",
        201,
        href,
        nested_role=nested_role,
    )

    # Permission check testing one more time
    try_action(user_helpless, ansible_bindings.RepositoriesAnsibleApi, "my_permissions", 404, href)


def test_ansible_distribution_rbac(
    ansible_bindings,
    ansible_repo_factory,
    ansible_distribution_factory,
    gen_users,
    try_action,
):
    user_creator, user_reader, user_helpless = gen_users(
        ["ansibledistribution", "ansiblerepository"]
    )

    # Create testing
    with user_creator:
        repo = ansible_repo_factory()
        distribution = ansible_distribution_factory(repo)
    try_action(
        user_reader, ansible_bindings.DistributionsAnsibleApi, "create", 403, gen_distribution()
    )
    try_action(
        user_helpless, ansible_bindings.DistributionsAnsibleApi, "create", 403, gen_distribution()
    )

    # List testing
    try_action(user_creator, ansible_bindings.DistributionsAnsibleApi, "list", 200)
    try_action(user_reader, ansible_bindings.DistributionsAnsibleApi, "list", 200)
    try_action(user_helpless, ansible_bindings.DistributionsAnsibleApi, "list", 200)

    # View testing
    view_args = [distribution.pulp_href]
    try_action(user_creator, ansible_bindings.DistributionsAnsibleApi, "read", 200, *view_args)
    try_action(user_reader, ansible_bindings.DistributionsAnsibleApi, "read", 200, *view_args)
    try_action(user_helpless, ansible_bindings.DistributionsAnsibleApi, "read", 404, *view_args)

    # Update testing
    update_args = [distribution.pulp_href, gen_distribution()]
    try_action(user_creator, ansible_bindings.DistributionsAnsibleApi, "update", 202, *update_args)
    try_action(user_reader, ansible_bindings.DistributionsAnsibleApi, "update", 403, *update_args)
    try_action(user_helpless, ansible_bindings.DistributionsAnsibleApi, "update", 404, *update_args)

    # Partial update testing
    partial_update_args = [distribution.pulp_href, {"name": str(uuid.uuid4())}]
    try_action(
        user_creator,
        ansible_bindings.DistributionsAnsibleApi,
        "partial_update",
        202,
        *partial_update_args,
    )
    try_action(
        user_reader,
        ansible_bindings.DistributionsAnsibleApi,
        "partial_update",
        403,
        *partial_update_args,
    )
    try_action(
        user_helpless,
        ansible_bindings.DistributionsAnsibleApi,
        "partial_update",
        404,
        *partial_update_args,
    )

    # Delete testing
    try_action(
        user_reader, ansible_bindings.DistributionsAnsibleApi, "delete", 403, distribution.pulp_href
    )
    try_action(
        user_helpless,
        ansible_bindings.DistributionsAnsibleApi,
        "delete",
        404,
        distribution.pulp_href,
    )
    try_action(
        user_creator,
        ansible_bindings.DistributionsAnsibleApi,
        "delete",
        202,
        distribution.pulp_href,
    )


def test_ansible_distribution_role_management(
    ansible_bindings,
    ansible_repo_factory,
    ansible_distribution_factory,
    gen_users,
    try_action,
):
    user_creator, user_reader, user_helpless = gen_users(
        ["ansibledistribution", "ansiblerepository"]
    )

    # Check role management apis
    with user_creator:
        repo = ansible_repo_factory()
        distribution = ansible_distribution_factory(repo)

    href = distribution.pulp_href

    # Permission check testing
    urperm_response = try_action(
        user_reader, ansible_bindings.DistributionsAnsibleApi, "my_permissions", 200, href
    )
    assert urperm_response.permissions == []
    ucperm_response = try_action(
        user_creator, ansible_bindings.DistributionsAnsibleApi, "my_permissions", 200, href
    )
    assert len(ucperm_response.permissions) > 0
    try_action(user_helpless, ansible_bindings.DistributionsAnsibleApi, "my_permissions", 404, href)

    # Add "creator" role testing
    nested_role = {"users": [user_reader.username], "role": "ansible.ansibledistribution_creator"}
    try_action(
        user_reader,
        ansible_bindings.DistributionsAnsibleApi,
        "add_role",
        403,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_helpless,
        ansible_bindings.DistributionsAnsibleApi,
        "add_role",
        404,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_creator,
        ansible_bindings.DistributionsAnsibleApi,
        "add_role",
        201,
        href,
        nested_role=nested_role,
    )

    # Permission check testing again
    urperm_response = try_action(
        user_reader, ansible_bindings.DistributionsAnsibleApi, "my_permissions", 200, href
    )
    assert len(urperm_response.permissions) == 1

    # Remove "viewer" role testing
    try_action(
        user_reader,
        ansible_bindings.DistributionsAnsibleApi,
        "remove_role",
        403,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_helpless,
        ansible_bindings.DistributionsAnsibleApi,
        "remove_role",
        404,
        href,
        nested_role=nested_role,
    )
    try_action(
        user_creator,
        ansible_bindings.DistributionsAnsibleApi,
        "remove_role",
        201,
        href,
        nested_role=nested_role,
    )

    # Permission check testing one more time
    urperm_response = try_action(
        user_reader, ansible_bindings.DistributionsAnsibleApi, "my_permissions", 200, href
    )
    assert len(urperm_response.permissions) == 0


REMOTES = {
    "collection": {"role_name": "collectionremote"},
    "git": {"role_name": "gitremote"},
    "role": {"role_name": "roleremote"},
}


@pytest.mark.parametrize("remote", REMOTES)
def test_remotes_rbac(
    ansible_bindings,
    gen_users,
    try_action,
    remote,
):
    REMOTE_APIS = {
        "collectionremote": ansible_bindings.RemotesCollectionApi,
        "gitremote": ansible_bindings.RemotesGitApi,
        "roleremote": ansible_bindings.RemotesRoleApi,
    }
    role_name = REMOTES[remote]["role_name"]
    remote_api = REMOTE_APIS[role_name]

    user_creator, user_reader, user_helpless = gen_users(role_name)

    # List testing
    try_action(user_creator, remote_api, "list", 200)
    try_action(user_reader, remote_api, "list", 200)
    try_action(user_helpless, remote_api, "list", 200)

    # Create testing
    remote = try_action(
        user_creator,
        remote_api,
        "create",
        201,
        gen_ansible_remote(url=ANSIBLE_GALAXY_URL),
    )
    try_action(
        user_reader,
        remote_api,
        "create",
        403,
        gen_ansible_remote(url=ANSIBLE_GALAXY_URL),
    )
    try_action(
        user_helpless,
        remote_api,
        "create",
        403,
        gen_ansible_remote(url=ANSIBLE_GALAXY_URL),
    )

    # View testing
    try_action(user_creator, remote_api, "read", 200, remote.pulp_href)
    try_action(user_reader, remote_api, "read", 200, remote.pulp_href)
    try_action(user_helpless, remote_api, "read", 404, remote.pulp_href)

    # Update testing
    update_args = [remote.pulp_href, gen_ansible_remote(url=ANSIBLE_GALAXY_URL)]
    try_action(user_creator, remote_api, "update", 202, *update_args)
    try_action(user_reader, remote_api, "update", 403, *update_args)
    try_action(user_helpless, remote_api, "update", 404, *update_args)

    # Partial update testing
    partial_update_args = [remote.pulp_href, {"name": str(uuid.uuid4())}]
    try_action(user_creator, remote_api, "partial_update", 202, *partial_update_args)
    try_action(user_reader, remote_api, "partial_update", 403, *partial_update_args)
    try_action(user_helpless, remote_api, "partial_update", 404, *partial_update_args)

    # Delete testing
    try_action(user_reader, remote_api, "delete", 403, remote.pulp_href)
    try_action(user_helpless, remote_api, "delete", 404, remote.pulp_href)
    try_action(user_creator, remote_api, "delete", 202, remote.pulp_href)


@pytest.mark.parametrize("remote", REMOTES)
def test_remotes_role_management(
    ansible_bindings,
    gen_users,
    try_action,
    remote,
):
    REMOTE_APIS = {
        "collectionremote": ansible_bindings.RemotesCollectionApi,
        "gitremote": ansible_bindings.RemotesGitApi,
        "roleremote": ansible_bindings.RemotesRoleApi,
    }
    role_name = REMOTES[remote]["role_name"]
    remote_api = REMOTE_APIS[role_name]

    user_creator, user_reader, user_helpless = gen_users(role_name)

    # Check role management apis
    remote = try_action(
        user_creator,
        remote_api,
        "create",
        201,
        gen_ansible_remote(url=ANSIBLE_GALAXY_URL),
    )
    href = remote.pulp_href

    # Permission check testing
    urperm_response = try_action(user_reader, remote_api, "my_permissions", 200, href)
    assert urperm_response.permissions == []
    ucperm_response = try_action(user_creator, remote_api, "my_permissions", 200, href)
    assert len(ucperm_response.permissions) > 0
    try_action(user_helpless, remote_api, "my_permissions", 404, href)

    # Add "viewer" role testing
    nested_role = {"users": [user_helpless.username], "role": f"ansible.{role_name}_viewer"}
    try_action(user_reader, remote_api, "add_role", 403, href, nested_role=nested_role)
    try_action(user_helpless, remote_api, "add_role", 404, href, nested_role=nested_role)
    try_action(user_creator, remote_api, "add_role", 201, href, nested_role=nested_role)

    # Permission check testing again
    uhperm_response = try_action(user_helpless, remote_api, "my_permissions", 200, href)
    assert len(uhperm_response.permissions) == 1

    # Remove "viewer" role testing
    try_action(user_reader, remote_api, "remove_role", 403, href, nested_role=nested_role)
    try_action(user_helpless, remote_api, "remove_role", 403, href, nested_role=nested_role)
    try_action(user_creator, remote_api, "remove_role", 201, href, nested_role=nested_role)

    # Permission check testing one more time
    try_action(user_helpless, remote_api, "my_permissions", 404, href)



import pytest
from pulpcore.client.pulp_ansible.exceptions import ApiException


@pytest.mark.parallel
def test_remote_with_url_only_is_allowed(ansible_collection_remote_factory):
    """Assert that a `CollectionRemote` with only a url can be created."""
    ansible_collection_remote_factory(url="https://example.com/")


@pytest.mark.parallel
def test_token_only_is_allowed(ansible_collection_remote_factory):
    """Assert that a `CollectionRemote` with `token` and no `auth_url` can be created."""
    ansible_collection_remote_factory(url="https://example.com/", token="this is a token string")


@pytest.mark.parallel
def test_update_auth_url(ansible_bindings, ansible_collection_remote_factory, monitor_task):
    """Assert that a `CollectionRemote` with `token` and no `auth_url` can be created."""
    remote = ansible_collection_remote_factory(
        url="https://example.com/",
        token="this is a token string",
        auth_url="https://example.com",
    )
    assert not hasattr(remote, "token")
    monitor_task(
        ansible_bindings.RemotesCollectionApi.partial_update(
            remote.pulp_href, {"auth_url": None}
        ).task
    )
    monitor_task(
        ansible_bindings.RemotesCollectionApi.partial_update(
            remote.pulp_href, {"auth_url": "https://example.com"}
        ).task
    )


@pytest.mark.parallel
def test_auth_url_requires_token(ansible_collection_remote_factory):
    """Assert that a `CollectionRemote` with `auth_url` and no `token` can't be created."""
    with pytest.raises(ApiException) as exc_info:
        ansible_collection_remote_factory(
            url="https://example.com/", auth_url="https://example.com"
        )
    assert exc_info.value.status == 400
    assert "When specifying 'auth_url' you must also specify 'token'." in exc_info.value.body


@pytest.mark.parallel
def test_remote_urls(ansible_collection_remote_factory):
    """This tests that the remote url ends with a "/"."""
    with pytest.raises(ApiException):
        ansible_collection_remote_factory(url="http://galaxy.ansible.com/api")

    with pytest.raises(ApiException):
        ansible_collection_remote_factory(url="http://galaxy.example.com")

    ansible_collection_remote_factory(url="https://galaxy.ansible.com")
    ansible_collection_remote_factory(url="https://galaxy.ansible.com/")



# Generated by Django 2.2.3 on 2019-08-19 14:38

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_increase_artifact_size_field'),
        ('ansible', '0008_collectionremote_requirements_file'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionImport',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('namespace', models.CharField(editable=False, max_length=64)),
                ('name', models.CharField(editable=False, max_length=64)),
                ('version', models.CharField(editable=False, max_length=32)),
                ('messages', django.contrib.postgres.fields.jsonb.JSONField(default=list, editable=False)),
                ('task', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='core.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import logging

from gettext import gettext as _


from pulpcore.plugin.models import ProgressReport
from pulpcore.plugin.stages import (
    DeclarativeVersion,
    Stage,
)
from pulp_ansible.app.models import AnsibleRepository, GitRemote
from pulp_ansible.app.tasks.collections import declarative_content_from_git_repo

log = logging.getLogger(__name__)


def synchronize(remote_pk, repository_pk, mirror=False):
    """
    Sync content from the remote repository.

    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.

    Raises:
        ValueError: If the remote does not specify a URL to sync.

    """
    remote = GitRemote.objects.get(pk=remote_pk)
    repository = AnsibleRepository.objects.get(pk=repository_pk)

    if not remote.url:
        raise ValueError(_("A remote must have a url specified to synchronize."))

    log.info(
        _("Synchronizing: repository=%(r)s remote=%(p)s"), {"r": repository.name, "p": remote.name}
    )
    first_stage = GitFirstStage(remote)
    d_version = DeclarativeVersion(first_stage, repository, mirror=mirror)
    return d_version.create()


class GitFirstStage(Stage):
    """
    The first stage of a pulp_ansible sync pipeline for git repositories.
    """

    def __init__(self, remote):
        """
        The first stage of a pulp_ansible sync pipeline.

        Args:
            remote (RoleRemote): The remote data to be used when syncing

        """
        super().__init__()
        self.remote = remote

        # Interpret download policy
        self.metadata_only = self.remote.metadata_only

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the ansible metadata.
        """
        async with ProgressReport(
            message="Cloning Git repository for Collection", code="sync.git.clone"
        ) as pb:
            d_content = await declarative_content_from_git_repo(
                self.remote, self.remote.url, self.remote.git_ref, self.metadata_only
            )
            await self.put(d_content)
            await pb.aincrement()



"""CRUD tests for AnsibleDistribution."""

import pytest
from collections import namedtuple

from pulpcore.client.pulp_ansible import ApiException


FakeRepository = namedtuple("FakeRepository", "pulp_href")


@pytest.mark.parallel
def test_crud_distribution(
    ansible_bindings,
    ansible_role_remote_factory,
    ansible_sync_factory,
    ansible_distribution_factory,
    monitor_task,
    has_pulp_plugin,
):
    repository = ansible_sync_factory(remote=ansible_role_remote_factory().pulp_href)
    distribution = ansible_distribution_factory(repository=repository)
    assert distribution.repository == repository.pulp_href
    assert distribution.repository_version is None

    with pytest.raises(ApiException) as exc_info:
        fake_repository = FakeRepository(
            pulp_href=repository.pulp_href[:-37] + "00000000-0000-0000-0000-000000000000/"
        )
        ansible_distribution_factory(repository=fake_repository)
    assert "Invalid hyperlink - Object does not exist." in exc_info.value.body

    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.DistributionsAnsibleApi.partial_update(
            distribution.pulp_href, {"repository": "this-is-invalid"}
        )
    assert "Invalid hyperlink - No URL match." in exc_info.value.body

    monitor_task(
        ansible_bindings.DistributionsAnsibleApi.partial_update(
            distribution.pulp_href, {"repository": None}
        ).task
    )
    monitor_task(
        ansible_bindings.DistributionsAnsibleApi.partial_update(
            distribution.pulp_href, {"repository_version": repository.latest_version_href}
        ).task
    )
    distribution = ansible_bindings.DistributionsAnsibleApi.read(distribution.pulp_href)
    assert distribution.repository is None
    assert distribution.repository_version == repository.latest_version_href

    if has_pulp_plugin("pulpcore", min="3.30.1"):
        # On older versions, this fails complaining repository and version cannot be used together.
        body = {
            "name": distribution.name,
            "base_path": distribution.base_path,
            "repository": repository.pulp_href,
        }
        monitor_task(
            ansible_bindings.DistributionsAnsibleApi.update(distribution.pulp_href, body).task
        )
        distribution = ansible_bindings.DistributionsAnsibleApi.read(distribution.pulp_href)
        assert distribution.repository == repository.pulp_href
        assert distribution.repository_version is None

    body = {
        "name": distribution.name,
        "base_path": distribution.base_path,
        "repository": repository.pulp_href,
        "repository_version": repository.latest_version_href,
    }
    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.DistributionsAnsibleApi.update(distribution.pulp_href, body)
    assert (
        "Only one of the attributes 'repository' and 'repository_version' "
        "may be used simultaneously."
    ) in exc_info.value.body

    monitor_task(ansible_bindings.DistributionsAnsibleApi.delete(distribution.pulp_href).task)
    with pytest.raises(ApiException) as exc_info:
        ansible_bindings.DistributionsAnsibleApi.delete(distribution.pulp_href)
    assert (
        "Not found." in exc_info.value.body
        or "No AnsibleDistribution matches the given query." in exc_info.value.body
    )



from urllib.parse import urljoin

from pulp_smash.pulp3.constants import (
    BASE_DISTRIBUTION_PATH,
    BASE_REPO_PATH,
)


GALAXY_ANSIBLE_BASE_URL = "https://galaxy.ansible.com"

ANSIBLE_ROLE_NAME = "ansible.role"

ANSIBLE_DISTRIBUTION_PATH = urljoin(BASE_DISTRIBUTION_PATH, "ansible/ansible/")

ANSIBLE_REPO_PATH = urljoin(BASE_REPO_PATH, "ansible/ansible/")

ANSIBLE_GALAXY_URL = urljoin(GALAXY_ANSIBLE_BASE_URL, "api/v1/roles/")

NAMESPACE_ANSIBLE = "?owner__username=ansible"

NAMESPACE_ELASTIC = "?owner__username=elastic"

NAMESPACE_PULP = "?owner__username=pulp"

NAMESPACE_TESTING = "?owner__username=testing"

ANSIBLE_FIXTURE_URL = urljoin(ANSIBLE_GALAXY_URL, NAMESPACE_ANSIBLE)

ANSIBLE_ELASTIC_FIXTURE_URL = urljoin(ANSIBLE_GALAXY_URL, NAMESPACE_ELASTIC)

ANSIBLE_ELASTIC_ROLE_NAMESPACE_NAME = "elastic.elasticsearch"

ANSIBLE_ELASTIC_ROLE = "elastic.elasticsearch,6.6.0"

ANSIBLE_FIXTURE_COUNT = 5

ANSIBLE_FIXTURE_CONTENT_SUMMARY = {ANSIBLE_ROLE_NAME: ANSIBLE_FIXTURE_COUNT}

ANSIBLE_DEMO_COLLECTION = "testing.k8s_demo_collection"
ANSIBLE_DEMO_COLLECTION_VERSION = "0.0.3"

ANSIBLE_DEMO_COLLECTION_REQUIREMENTS = f"collections:\n  - {ANSIBLE_DEMO_COLLECTION}"

COLLECTION_METADATA = {"name": "k8s_demo_collection", "version": "0.0.3"}
"""Metadata was extracted from
https://galaxy.ansible.com/api/v2/collections/testing/k8s_demo_collection/versions/0.0.3/"""

ANSIBLE_COLLECTION_FILE_NAME = "testing-k8s_demo_collection-0.0.3.tar.gz"

ANSIBLE_COLLECTION_UPLOAD_FIXTURE_URL = urljoin(
    GALAXY_ANSIBLE_BASE_URL, f"download/{ANSIBLE_COLLECTION_FILE_NAME}"
)

TEST_COLLECTION_CONFIGS = [
    {"name": "a", "namespace": "test", "dependencies": {"test.b": "*"}},
    {"name": "b", "namespace": "test", "dependencies": {"test.c": "*"}},
    {"name": "c", "namespace": "test", "dependencies": {"test.d": "*"}},
    {"name": "d", "namespace": "test"},
    {"name": "e", "namespace": "test", "dependencies": {"test.f": "*", "test.g": "*"}},
    {"name": "f", "namespace": "test", "dependencies": {"test.h": "<=3.0.0"}},
    {"name": "g", "namespace": "test", "dependencies": {"test.d": "*"}},
    {"name": "h", "namespace": "test", "version": "1.0.0"},
    {"name": "h", "namespace": "test", "version": "2.0.0"},
    {"name": "h", "namespace": "test", "version": "3.0.0"},
    {"name": "h", "namespace": "test", "version": "4.0.0"},
    {"name": "h", "namespace": "test", "version": "5.0.0"},
]



# Generated by Django 2.2.4 on 2019-08-19 19:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0009_collectionimport'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='ansibledistribution',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='ansibleremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='collectionremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='collectionversion',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='role',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterField(
            model_name='ansibledistribution',
            name='basedistribution_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_ansibledistribution', serialize=False, to='core.BaseDistribution'),
        ),
        migrations.AlterField(
            model_name='ansibledistribution',
            name='repository',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='ansible_ansibledistribution', to='core.Repository'),
        ),
        migrations.AlterField(
            model_name='ansibledistribution',
            name='repository_version',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='ansible_ansibledistribution', to='core.RepositoryVersion'),
        ),
        migrations.AlterField(
            model_name='ansibleremote',
            name='remote_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_ansibleremote', serialize=False, to='core.Remote'),
        ),
        migrations.AlterField(
            model_name='collectionremote',
            name='remote_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_collectionremote', serialize=False, to='core.Remote'),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_collectionversion', serialize=False, to='core.Content'),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='tags',
            field=models.ManyToManyField(editable=False, related_name='ansible_collectionversion', to='ansible.Tag'),
        ),
        migrations.AlterField(
            model_name='role',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_role', serialize=False, to='core.Content'),
        ),
    ]



import pytest
import hashlib
import requests
import random
import string

import numpy as np
from PIL import Image


def random_string(n=10):
    """Create a random lower_case + digits string."""
    a = random.choice(string.ascii_lowercase)
    return a + "".join(random.choices(string.ascii_lowercase + string.digits, k=n - 1))


@pytest.fixture
def ansible_repo_and_distro_factory(ansible_repo_factory, ansible_distribution_factory):
    """Factory to produce a repo and distro."""

    def _factory(repo_kwargs=None):
        repo_kwargs = repo_kwargs or {}
        ansible_repo = ansible_repo_factory(**repo_kwargs)
        distro = ansible_distribution_factory(ansible_repo)
        return ansible_repo, distro

    return _factory


@pytest.fixture
def random_image_factory(tmp_path):
    """Factory to produce a random 100x100 image."""

    def _random_image():
        imarray = np.random.rand(100, 100, 3) * 255
        im = Image.fromarray(imarray.astype("uint8")).convert("RGBA")
        path = tmp_path / f"{random_string()}.png"
        im.save(path)
        return path

    return _random_image


def test_crud_namespaces(
    ansible_bindings,
    ansible_repo_and_distro_factory,
    monitor_task,
):
    """Test Basic Creating, Reading, Updating and 'Deleting' operations on Namespaces."""
    repo, distro = ansible_repo_and_distro_factory()
    kwargs = {"path": distro.base_path, "distro_base_path": distro.base_path}

    links = [
        {"name": "link1", "url": "http://galaxy.ansible.com"},
        {"name": "link2", "url": "http://begta-galaxy.ansible.com"},
    ]

    # Test Basic Creation
    name = random_string()
    task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.create(
        name=name, description="hello", company="Testing Co.", links=links, **kwargs
    )
    result = monitor_task(task.task)

    assert len(result.created_resources) == 2
    namespace_href = result.created_resources[1]
    assert "content/ansible/namespaces/" in namespace_href

    def _link_to_dict(resp):
        return {x.name: x.url for x in resp}

    link_comparison = {x["name"]: x["url"] for x in links}

    # Test Reading Namespace from Pulp API
    namespace = ansible_bindings.ContentNamespacesApi.read(namespace_href)
    assert namespace.name == name
    assert namespace.company == "Testing Co."
    assert namespace.description == "hello"
    assert namespace.metadata_sha256 != ""
    assert namespace.avatar_url is None
    assert _link_to_dict(namespace.links) == link_comparison

    # Test Reading Namespace from Galaxy API
    v3_namespace = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.read(
        name=name, **kwargs
    )
    assert v3_namespace.pulp_href == namespace.pulp_href
    assert v3_namespace.name == name
    assert v3_namespace.company == "Testing Co."
    assert v3_namespace.description == "hello"
    assert v3_namespace.metadata_sha256 != ""
    assert v3_namespace.avatar_url is None
    assert _link_to_dict(v3_namespace.links) == link_comparison

    v3_list = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.list(**kwargs)
    assert v3_list.count == 1

    # Test Update Namespace (Creates a new Namespace w/ updated metadata)
    task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.partial_update(
        name=name, description="hey!", **kwargs
    )
    result = monitor_task(task.task)
    assert len(result.created_resources) == 2
    repo_version = result.created_resources[0]
    assert repo_version[-2] == "2"
    updated_namespace_href = result.created_resources[1]
    assert updated_namespace_href != namespace_href

    updated_namespace = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.read(
        name=name, **kwargs
    )
    assert updated_namespace.pulp_href == updated_namespace_href
    assert updated_namespace.name == name
    assert updated_namespace.company == "Testing Co."
    assert updated_namespace.description == "hey!"
    assert updated_namespace.metadata_sha256 != v3_namespace.metadata_sha256

    # Test Updating Namespace back to original description brings back original object
    task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.partial_update(
        name=name, description="hello", links=links, **kwargs
    )
    result = monitor_task(task.task)
    assert len(result.created_resources) == 2
    repo_version = result.created_resources[0]
    assert repo_version[-2] == "3"
    updated2_namespace_href = result.created_resources[1]
    assert updated2_namespace_href == namespace_href

    # Test 'Deleting' a Namespace (Namespace are removed from Galaxy API, but are still in Pulp)
    task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.delete(
        name=name, **kwargs
    )
    result = monitor_task(task.task)
    assert len(result.created_resources) == 1
    repo_version = result.created_resources[0]
    assert repo_version[-2] == "4"

    namespaces = ansible_bindings.ContentNamespacesApi.list(name=name)
    assert namespaces.count == 2

    v3_namespaces = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.list(
        **kwargs
    )
    assert v3_namespaces.count == 0


def test_namespace_avatar(
    ansible_bindings,
    ansible_repo_and_distro_factory,
    random_image_factory,
    monitor_task,
):
    """Test creating a Namespace w/ an Avatar & downloading the image from the Galaxy API."""
    repo, distro = ansible_repo_and_distro_factory()
    kwargs = {"path": distro.base_path, "distro_base_path": distro.base_path}

    name = random_string()
    avatar_path = random_image_factory()
    with open(avatar_path, "rb") as av:
        avatar_sha256 = hashlib.sha256(av.read()).hexdigest()

    task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.create(
        name=name, avatar=avatar_path, **kwargs
    )
    result = monitor_task(task.task)

    assert len(result.created_resources) == 2
    namespace_href = result.created_resources[1]

    namespace = ansible_bindings.ContentNamespacesApi.read(namespace_href)
    assert namespace.avatar_sha256 == avatar_sha256
    assert namespace.avatar_url.endswith(f"{namespace_href}avatar/")

    v3_namespace = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.read(
        name=name, **kwargs
    )
    assert v3_namespace.pulp_href == namespace_href
    assert v3_namespace.avatar_sha256 == avatar_sha256
    assert v3_namespace.avatar_url.endswith(f"{namespace_href}avatar/")

    r = requests.get(v3_namespace.avatar_url)
    assert r.status_code == 200
    downloaded_sha256 = hashlib.sha256(r.content).hexdigest()
    assert downloaded_sha256 == v3_namespace.avatar_sha256


def test_namespace_syncing(
    ansible_bindings,
    ansible_repo_and_distro_factory,
    build_and_upload_collection,
    random_image_factory,
    monitor_task,
    ansible_sync_factory,
    ansible_collection_remote_factory,
    pulp_admin_user,
):
    """Test syncing a Collection w/ a Namespace also syncs the Namespace."""
    # Set up first Repo to have 3 Collections and 3 Namespaces
    repo1, distro1 = ansible_repo_and_distro_factory()
    kwargs1 = {"path": distro1.base_path, "distro_base_path": distro1.base_path}

    collections = []
    namespaces = {}
    for i in range(3):
        namespace = random_string()
        collection, _ = build_and_upload_collection(repo1, config={"namespace": namespace})
        avatar_path = random_image_factory()
        task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.create(
            name=namespace, avatar=avatar_path, **kwargs1
        )
        result = monitor_task(task.task)
        collections.append(collection)
        namespaces[namespace] = result.created_resources[-1]

    # Set up second Repo and sync 2 Collections from first Repo
    repo2, distro2 = ansible_repo_and_distro_factory()
    kwargs2 = {"path": distro2.base_path, "distro_base_path": distro2.base_path}
    collections_string = "\n  - ".join((f"{c.namespace}.{c.name}" for c in collections[0:2]))
    requirements = f"collections:\n  - {collections_string}"
    remote = ansible_collection_remote_factory(
        url=distro1.client_url,
        requirements_file=requirements,
        username=pulp_admin_user.username,
        password=pulp_admin_user.password,
    )

    ansible_sync_factory(ansible_repo=repo2, remote=remote.pulp_href)
    # 2 Namespaces should have also been synced
    synced_namespaces = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.list(
        **kwargs2
    )
    assert synced_namespaces.count == 2
    for namespace in synced_namespaces.results:
        assert namespace.name in namespaces
        assert namespaces[namespace.name] == namespace.pulp_href
